"""
# TCP Stream Module
This module provide an api class for managing tcp connection

## Class
- `TcpStream` : a class for managing tcp connection
"""

import socket
import select
import logging

from inovopy.socket import SocketException, EndOfCommunication, detect_interfaces
from inovopy.util import clean
from inovopy.util import Loggable


class TcpStream(Loggable):
    """
    # TcpStream
    A class for managing tcp connetion

    ## Usage
    ```python
    from inovopy.socket import TcpListener, TcpStream

    example_listener : TcpListener = TcpListener()

    example_stream : TcpStream = example_listener.accept()

    example_stream.write("send a message")

    read_message = example_stream.read()
    ```
    """

    def __init__(self, conn: socket.socket, logger: logging.Logger | str | None = None):
        """
        initalize the stream

        ## Parameter
        - `conn : socket.socket` : the socket connection
        - `logger: logging.Logger | str | None` :
            - if `logger` is instance of `logging.Logger`, it will log with it;
            - if `logger` is `str`, new logger will be created with it as name
            - otherwise, no log
        """
        super().__init__(logger)
        self.__conn: socket.socket = conn

    @classmethod
    def connect(
        cls,
        ip: str | None = None,
        port: int = 50003,
        logger: logging.Logger | str | None = None,
    ) -> "TcpStream":
        """
        Try to connect to an socket address

        ## Parameter
        - `ip : str`: if `None`, will try local machine
        - `port: int` : port to connect to
        - `logger: Logger`: logger for the resulted `TcpStream`

        ## Return
        `TcpStream` the resulted connection

        ## Exception:
        `SocketException`
        - if no network interface found
        """
        logger = TcpStream.get_class_logger()

        if not ip:
            ip = "localhost"

        if ip == "localhost":
            ips = detect_interfaces()
            if len(ips) == 0:
                logger.error("No interface found")
                raise SocketException(
                    "No Network Interfaces, please make sure device is connected to network"
                )

        logger.debug("Start connecting . . .")
        try:
            conn = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            conn.connect((ip, port))
            logger.debug(f"Successfully connected to {ip}:{port}")
            return TcpStream(conn=conn, logger=logger)
        except OSError as e:
            logger.warning("Failed to connect")
            logger.warning(f"{e}")
            raise SocketException("Failed to connect") from e

    @property
    def ip(self) -> str:
        """get the local ip of the socket"""
        return self.__conn.getsockname()[0]

    def try_read(self, timeout: float = 1.0) -> str | None:
        """
        try to read a message from the socket
        ## Parameter
        - `timout : float = 1.0` : time out for the reading

        ## Return
        `str` : read message; or
        `None` : if no message is recived
        """
        ready_socket, _, _ = select.select([self.__conn], [], [], timeout)
        if not ready_socket:
            return None
        return self.read()

    def read(self) -> str:
        """
        read a message form the connection

        maximum byte : `4096`

        ## Return:
        `str` the message read

        ## Exception:
        `SocketException`:
        - if the read failed
        - if `EOF` character is read
        """
        try:
            self.__conn.setblocking(True)
            data = self.__conn.recv(4096)
        except OSError as e:
            self.logger.error("Error occur during socket read!")
            self.logger.error(f"{e}")
            raise SocketException("Read Error") from e

        if not data:
            self.__conn.close()
            self.logger.error("Read Error : EOF detected")
            raise EndOfCommunication()

        s = clean(data.decode("UTF-8"))
        self.logger.debug(f"....{s}")

        return s

    def write(self, msg: str):
        """
        write a message to the connection

        ## Parameter:
        - `msg : str` : string message to send
        """
        try:
            self.logger.debug(f"{msg}")
            msg = f"{msg}\n"
            self.__conn.send(msg.encode())
        except Exception as e:
            self.logger.error("Error occur during socket write!")
            self.logger.error(f"{e}")
            raise SocketException("Write Error") from e

    def __del__(self):
        try:
            self.logger.debug("closing socket")
            self.__conn.close()
        except OSError as e:
            try:
                self.logger.error("Failed to shut down socket!")
                self.logger.error(f"{e}")
            except OSError:
                pass



"""
# Ros Bridge Module
This module contain function and class for using rosbridge
api to command and get state from inovo arm

`inovopy.rosbridge.InovoRos`
"""

from contextlib import contextmanager
import logging
import typing
from enum import Enum

import inovopy.util
import roslibpy

# Topic
TOPIC_TCP_SPEED = "/default_move_group/tcp_speed"
TOPIC_TCP_POSE = "/default_move_group/tcp_pose"
TOPIC_JOINT_STATE = "/robot/joint_states"
TOPIC_POWER_STATUS = "/psu/status"
TOPIC_ROBOT_STATUS = "/robot/robot_state"
TOPIC_ESTOP_STATUS = "/psu/estop/state"
TOPIC_SAFE_STOP_STATUS = "/psu/safe_stop/state"
TOPIC_RUNTIME_STATE = "/sequence/runtime_state"
TOPIC_ARM_STATE = "/robot/arm_state"


TYPE_TCP_SPEED = "commander_msgs/SpeedStamped"
TYPE_TCP_POSE = "geometry_msgs/PoseStamped"
TYPE_JOINT_STATE = "sensor_msgs/JointState"
TYPE_POWER_STATUS = "psu_msgs/Status"
TYPE_ROBOT_STATUS = "arm_msgs/RobotState"
TYPE_ESTOP_STATUS = "psu_msgs/SafetyCircuitState"
TYPE_SAFE_STOP_STATUS = "psu_msgs/SafetyCircuitState"
TYPE_RUNTIME_STATE = "commander_msgs/RuntimeState"
TYPE_ARM_STATE = "arm_msgs/ArmState"

# Service
SERVICE_SAFE_STOP_RESET = "/psu/safe_stop/reset"
SERVICE_ESTOP_RESET = "/psu/estop/reset"
SERVICE_POWER_ON = "/psu/enable"
SERVICE_POWER_OFF = "/psu/disable"
SERVICE_ROBOT_ENABLE = "/robot/enable"
SERVICE_ROBOT_DISABLE = "/robot/disable"

SERVICE_SEQUENCE_START = "/sequence/start"
SERVICE_SEQUENCE_STOP = "/sequence/stop"
SERVICE_SEQUENCE_PAUSE = "/sequence/pause"
SERVICE_SEQUENCE_STEP = "/sequence/step"
SERVICE_SEQUENCE_DEBUG = "/sequence/debug"
SERVICE_SEQUENCE_CONTINUE = "/sequence/continue"
SERVICE_SEQUENCE_SET_VAR = "/sequence/set_var"
SERVICE_SEQUENCE_GET_VAR = "/sequence/get_var"


TYPE_TRIGGER = "std_srvs/Trigger"
TYPE_RUN_SEQUENCE = "commander_msgs/RunSequence"
TYPE_SET_VAR = "commander_msgs/set_var"
TYPE_GET_VAR = "commander_msgs/get_var"

# Publish Service
TOPIC_CARTESIAN_JOG = "/default_move_group/cartesian_jog"
TYPE_CARTESIAN_JOG_DEMAND = "commander_msgs/CartesianJogDemand"


class RuntimeState(Enum):
    Idle = 0
    Running = 1
    Paused = 2
    PausedOnError = 3


class Variable:
    name: str
    dtype: str
    value: str

    def __init__(self, name: str, dtype: str, value: str):
        self.name = name
        self.dtype = dtype
        self.value = value

    @classmethod
    def from_message(cls, message) -> "Variable":
        return Variable(message["name"], message["type"], message["value"])


class JointState:
    age: int = 0
    current: float = 0
    drive_temp: float = 0
    ff_torque: float = 0

    joint_temp: float = 0
    motor_temp: float = 0
    output_gain: float = 0
    position: float = 0

    state: int = 0
    status: int = 0

    target_position: float = 0
    torque: float = 0
    velocity: float = 0

    @classmethod
    def attr_list(cls) -> list[str]:
        return [
            "age",
            "current",
            "drive_temp",
            "ff_torque",
            "joint_temp",
            "motor_temp",
            "output_gain",
            "position",
            "state",
            "status",
            "target_position",
            "torque",
            "velocity",
        ]

    def __init__(self):
        pass

    @classmethod
    def from_message(cls, message) -> "JointState":
        js = JointState()

        for a in JointState.attr_list():
            setattr(js, a, message[a])

        return js

    def __str__(self):
        return str(dict((a, getattr(self, a)) for a in JointState.attr_list()))


class InovoServiceResponse:
    """
    Response when calling service

    Field:
    - `success : bool` : whether the service call success
    - `message : str` : message from service call, usually error message
    - `value : str | None` : extra information from response, used in `runtime_get_var`
    """

    success: bool
    message: str
    value: str | None

    def __init__(self, message: dict):
        self.success = message["success"]
        self.message = message["message"]
        self.value = message.get("value")

    def __str__(self):
        return str(
            {"success": self.success, "message": self.message, "value": self.value}
        )

    @classmethod
    def from_message(cls, message) -> "InovoServiceResponse":
        return InovoServiceResponse(message)


class InovoRos(inovopy.util.Loggable):
    """
    # RosBridge
    A class managing rosbridge api communication

    ## Usage
    ```python
    from inovopy.rosbridge import InovoRos

    ros = InovoRos("192.168.1.1")

    ros.estop_reset()
    ```
    """

    # Ros
    ros: roslibpy.Ros
    """connection manager to ROS server (`roslibpy.Ros`)"""

    # Tcp Speed
    tcp_speed_lin: float = 0
    tcp_speed_ang: float = 0

    # Tcp Pose
    tcp_pose_vec: tuple[float, float, float] = (0, 0, 0)
    tcp_pose_quat: tuple[float, float, float, float] = (0, 0, 0, 0)

    # Joint State
    joint_pos: list[float] = []
    joint_vel: list[float] = []
    joint_eff: list[float] = []

    # Power Status
    voltage: float = 0
    current: float = 0
    power_status: str = 0

    # Robot Status
    driver_state: str = ""
    drive_powered: bool = False

    # EStop Status
    estop_active: bool = False
    estop_circuit: bool = False

    # Safe Stop Status
    safe_stop_active: bool = False
    safe_stop_circuit: bool = False

    # Runtime State
    active_blocks: list[str] = []
    current_block_progress: float = 0
    runtime_status: RuntimeState = 0
    variables: list[Variable] = []

    # Arm State
    enable: bool = False
    state: int = 0
    joint_states: list[JointState] = []

    def __init__(self, host: str, logger: logging.Logger | str | None = None):
        """
        initalize `InovoRos`

        ## Args
        - `host : str` : host of psu, preferably in form of `192.168.x.x`
        - `logger: logging.Logger | str | None` :
            - if `logger` is instance of `logging.Logger`, it will log with it;
            - if `logger` is `str`, new logger will be created with it as name
            - otherwise, no log
        """
        super().__init__(logger)

        self.host: str = host

        self.info(f"ros bridge initalized with host : {self.host}")
        if not "192.168" in self.host:
            self.warning("host is not in form of 192.168.x.x")
            self.warning("this might cause networking latency")

        self.ros = roslibpy.Ros(host, 9090)
        self.ros.run()

        build_list = [
            (self.__tcp_speed, TOPIC_TCP_SPEED, TYPE_TCP_SPEED),
            (self.__tcp_pose, TOPIC_TCP_POSE, TYPE_TCP_POSE),
            (self.__joint_state, TOPIC_JOINT_STATE, TYPE_JOINT_STATE),
            (self.__power_status, TOPIC_POWER_STATUS, TYPE_POWER_STATUS),
            (self.__robot_status, TOPIC_ROBOT_STATUS, TYPE_ROBOT_STATUS),
            (self.__estop_status, TOPIC_ESTOP_STATUS, TYPE_ESTOP_STATUS),
            (self.__safe_stop_status, TOPIC_SAFE_STOP_STATUS, TYPE_SAFE_STOP_STATUS),
            (self.__runtime_state, TOPIC_RUNTIME_STATE, TYPE_RUNTIME_STATE),
            (self.__arm_state, TOPIC_ARM_STATE, TYPE_ARM_STATE),
        ]

        self.subscribers: dict[str, roslibpy.Topic] = {}

        for cb, name, message_type in build_list:
            self.info(f"Subscribing to topic: {name}, type : {message_type}")
            sub = roslibpy.Topic(self.ros, name, message_type)
            sub.subscribe(cb)
            self.subscribers[name] = sub

    # Topic Call Back
    def __tcp_speed(self, message):
        self.tcp_speed_lin = message["speed"]["linear"]
        self.tcp_speed_ang = message["speed"]["angular"]

    def __tcp_pose(self, message):
        vec = message["pose"]["position"]
        self.tcp_pose_vec = (vec["x"], vec["y"], vec["z"])

        quat = message["pose"]["orientation"]
        self.tcp_pose_quat = (
            quat["x"],
            quat["y"],
            quat["z"],
            quat["w"],
        )

    def __joint_state(self, message):
        self.joint_pos = message["position"]
        self.joint_vel = message["velocity"]
        self.joint_eff = message["effort"]

    def __power_status(self, message):
        self.power_status = message["state"]
        self.voltage = message["voltage"]
        self.current = message["current"]

    def __robot_status(self, message):
        self.driver_state = message["driver_state"]
        self.drives_powered = message["drives_powered"]

    def __estop_status(self, message):
        self.estop_active = message["active"]
        self.estop_circuit = message["circuit_complete"]

    def __safe_stop_status(self, message):
        self.safe_stop_active = message["active"]
        self.safe_stop_circuit = message["circuit_complete"]

    def __runtime_state(self, message):
        self.active_blocks = message["active_blocks"]
        self.current_block_progress = message["current_block_progress"]
        self.runtime_status = RuntimeState(message["state"])
        self.variables = []

        for v in message["variables"]:
            self.variables.append(Variable.from_message(v))

    def __arm_state(self, message):
        self.enabled = message["enabled"]
        self.state = message["state"]
        self.joint_states = []

        for j in message["joint_states"]:
            self.joint_states.append(JointState.from_message(j))

    # Service

    # PSU/Robot
    def safe_stop_reset(self) -> InovoServiceResponse:
        req = roslibpy.ServiceRequest()
        res = roslibpy.Service(self.ros, SERVICE_SAFE_STOP_RESET, TYPE_TRIGGER).call(
            req
        )
        return InovoServiceResponse.from_message(res)

    def estop_reset(self) -> InovoServiceResponse:
        req = roslibpy.ServiceRequest()
        res = roslibpy.Service(self.ros, SERVICE_ESTOP_RESET, TYPE_TRIGGER).call(req)
        return InovoServiceResponse.from_message(res)

    def power_on(self) -> InovoServiceResponse:
        req = roslibpy.ServiceRequest()
        res = roslibpy.Service(self.ros, SERVICE_POWER_ON, TYPE_TRIGGER).call(req)
        return InovoServiceResponse.from_message(res)

    def power_off(self) -> InovoServiceResponse:
        req = roslibpy.ServiceRequest()
        res = roslibpy.Service(self.ros, SERVICE_POWER_OFF, TYPE_TRIGGER).call(req)
        return InovoServiceResponse.from_message(res)

    def robot_enable(self) -> InovoServiceResponse:
        req = roslibpy.ServiceRequest()
        res = roslibpy.Service(self.ros, SERVICE_ROBOT_ENABLE, TYPE_TRIGGER).call(req)
        return InovoServiceResponse.from_message(res)

    def robot_disable(self) -> InovoServiceResponse:
        req = roslibpy.ServiceRequest()
        res = roslibpy.Service(self.ros, SERVICE_ROBOT_DISABLE, TYPE_TRIGGER).call(req)
        return InovoServiceResponse.from_message(res)

    # Runtime
    def runtime_start(self, procedure_name: str | None = None) -> InovoServiceResponse:
        """
        start a sequence in runtime

        ## Args:
        - `procedure_name : str | None`:
            - if `None`, will start from start block
            - if `str`, will try to start function with respected name
        """

        payload = {}

        if procedure_name:
            payload["procedure_name"] = procedure_name

        req = roslibpy.ServiceRequest(payload)
        res = roslibpy.Service(
            self.ros, SERVICE_SEQUENCE_START, TYPE_RUN_SEQUENCE
        ).call(req)
        return InovoServiceResponse.from_message(res)

    def runtime_stop(self) -> InovoServiceResponse:
        req = roslibpy.ServiceRequest()
        res = roslibpy.Service(self.ros, SERVICE_SEQUENCE_STOP, TYPE_TRIGGER).call(req)
        return InovoServiceResponse.from_message(res)

    def runtime_pause(self) -> InovoServiceResponse:
        req = roslibpy.ServiceRequest()
        res = roslibpy.Service(self.ros, SERVICE_SEQUENCE_PAUSE, TYPE_TRIGGER).call(req)
        return InovoServiceResponse.from_message(res)

    def runtime_step(self) -> InovoServiceResponse:
        req = roslibpy.ServiceRequest()
        res = roslibpy.Service(self.ros, SERVICE_SEQUENCE_STEP, TYPE_TRIGGER).call(req)
        return InovoServiceResponse.from_message(res)

    def runtime_debug(self) -> InovoServiceResponse:
        req = roslibpy.ServiceRequest()
        res = roslibpy.Service(self.ros, SERVICE_SEQUENCE_DEBUG, TYPE_TRIGGER).call(req)
        return InovoServiceResponse.from_message(res)

    def runtime_continue(self) -> InovoServiceResponse:
        req = roslibpy.ServiceRequest()
        res = roslibpy.Service(self.ros, SERVICE_SEQUENCE_CONTINUE, TYPE_TRIGGER).call(
            req
        )
        return InovoServiceResponse.from_message(res)

    def runtime_get_var(self, name: str) -> InovoServiceResponse:
        req = roslibpy.ServiceRequest({"name": name})
        res = roslibpy.Service(self.ros, SERVICE_SEQUENCE_GET_VAR, TYPE_GET_VAR).call(
            req
        )
        return InovoServiceResponse.from_message(res)

    def runtime_set_var(self, name: str, value: typing.Any) -> InovoServiceResponse:
        req = roslibpy.ServiceRequest({"name": name, "value": str(value)})
        res = roslibpy.Service(self.ros, SERVICE_SEQUENCE_SET_VAR, TYPE_SET_VAR).call(
            req
        )
        return InovoServiceResponse.from_message(res)

    @contextmanager
    def jog(self):
        """
        Enter a context of jog command mode. Advertise when enter context
        and automatically unadvertise when exit.

        ## Yields:
        - `JogPublisher`: api for jog command

        ## Example:
        ```python
        import time
        from inovopy.rosbridge import InovoRos

        ros = InovoRos("192.168.1.122", "Jog")

        with ros.jog() as j:
            for _ in range(1000):
                j.jog(z=-0.01)
                time.sleep(0.01)
            for _ in range(1000):
                j.jog(z=0.01)
                time.sleep(0.01)
        ```
        """

        top = roslibpy.Topic(self.ros, TOPIC_CARTESIAN_JOG, TYPE_CARTESIAN_JOG_DEMAND)
        top.advertise()

        publisher = JogPublisher(top)

        yield publisher

        publisher.publisher.unadvertise()
        del publisher

    # Destruction
    def unsubscribe_all(self):
        for n, sub in self.subscribers.items():
            self.info(f"unsubscribing topic : {n}")
            sub.unsubscribe()

    def resubscribe_all(self):
        for n, sub in self.subscribers.items():
            self.info(f"resubscribing topic : {n}")
            sub.subscribe()

    def __del__(self):
        self.ros.close()


class JogPublisher:
    """
    context manager for jog

    automatically unadvertise when `del`
    """

    publisher: roslibpy.Topic

    def __init__(self, publisher: roslibpy.Topic):
        self.publisher = publisher

        self.publisher.advertise()

    def jog(
        self,
        x: float = 0,
        y: float = 0,
        z: float = 0,
        rx: float = 0,
        ry: float = 0,
        rz: float = 0,
    ):
        """
        Send a cartesian jog twist demand

        Args:
        - `x : float`: target linear velocity x. Defaults to 0.
        - `y : float`: target linear velocity y. Defaults to 0.
        - `z : float`: target linear velocity z. Defaults to 0.
        - `rx : float`: target angular velocity rx. Defaults to 0.
        - `ry : float`: target angular velocity ry. Defaults to 0.
        - `rz : float`: target angular velocity ry. Defaults to 0.
        """
        self.publisher.publish(
            roslibpy.Message(
                {
                    "twist": {
                        "linear": {
                            "x": x,
                            "y": y,
                            "z": z,
                        },
                        "angular": {
                            "x": rx,
                            "y": ry,
                            "z": rz,
                        },
                    }
                }
            )
        )

    def __del__(self):
        self.publisher.unadvertise()



"""
# Transform Module
This module provide useful class and function for spatial calculation

## Class
- `Transform` : A spatial transform represented a translation and rotation

## Function
- `deg_to_rad` : translate degree to radian
- `rad_to_deg` : translate radian to degree
- `rx_mat` : compute rotation matrx of rotation along x axis
- `ry_mat` : compute rotation matrx of rotation along y axis
- `rz_mat` : compute rotation matrx of rotation along z axis
- `euler_to_mat` : translate euler angle to rotation matrix
- `mat_to_euler` : translate rotation matrix to euler angle
"""

from typing import Tuple
import re

import numpy
import numpy.typing as numpy_typing

from inovopy.geometry import IntoRobotCommand, deg_to_rad, rad_to_deg


def rx_mat(deg: float) -> numpy_typing.NDArray[numpy.float64]:
    """compute rotation matrx of rotation along x axis"""
    mat = numpy.eye(3)
    mat[1, 1] = numpy.cos(deg_to_rad(deg))
    mat[1, 2] = -numpy.sin(deg_to_rad(deg))
    mat[2, 2] = numpy.cos(deg_to_rad(deg))
    mat[2, 1] = numpy.sin(deg_to_rad(deg))
    return mat


def ry_mat(deg: float) -> numpy_typing.NDArray[numpy.float64]:
    """compute rotation matrx of rotation along y axis"""
    mat = numpy.eye(3)
    mat[0, 0] = numpy.cos(deg_to_rad(deg))
    mat[0, 2] = numpy.sin(deg_to_rad(deg))
    mat[2, 2] = numpy.cos(deg_to_rad(deg))
    mat[2, 0] = -numpy.sin(deg_to_rad(deg))
    return mat


def rz_mat(deg: float) -> numpy_typing.NDArray[numpy.float64]:
    """compute rotation matrx of rotation along z axis"""
    mat = numpy.eye(3)
    mat[0, 0] = numpy.cos(deg_to_rad(deg))
    mat[0, 1] = -numpy.sin(deg_to_rad(deg))
    mat[1, 1] = numpy.cos(deg_to_rad(deg))
    mat[1, 0] = numpy.sin(deg_to_rad(deg))
    return mat


def euler_to_mat(
    euler_deg: Tuple[float, float, float],
) -> numpy_typing.NDArray[numpy.float64]:
    """translate euler angle to rotation matrix"""
    rx = rx_mat(euler_deg[0])
    ry = ry_mat(euler_deg[1])
    rz = rz_mat(euler_deg[2])
    return numpy.matmul(rz, numpy.matmul(ry, rx))


def mat_to_euler(
    mat: numpy_typing.NDArray[numpy.float64],
) -> Tuple[float, float, float]:
    """translate rotation matrix to euler angle"""
    r31 = mat[2, 0]
    if numpy.abs(r31) != 1:
        ry = -numpy.arcsin(r31)
        rx = numpy.arctan2(mat[2, 1] / numpy.cos(ry), mat[2, 2] / numpy.cos(ry))
        rz = numpy.arctan2(mat[1, 0] / numpy.cos(ry), mat[0, 0] / numpy.cos(ry))
    else:
        rz = 0
        if r31 == -1:
            ry = numpy.pi / 2
            rx = numpy.arctan2(mat[0, 1], mat[0, 2])
        else:
            ry = -numpy.pi / 2
            rx = numpy.arctan2(-mat[0, 1], -mat[0, 2])
    return (rad_to_deg(rx), rad_to_deg(ry), rad_to_deg(rz))


class Transform(IntoRobotCommand):
    """
    # Transform
    A class representing spatial transform,
    compose of a translation and a rotation

    ## Representation
    - `vec_mm` : a 3D vector with unit in mm
    - `euler_deg` : a set of euler angle wiht unit in degree
    """

    def __init__(
        self,
        vec_mm: Tuple[float, float, float] = (0, 0, 0),
        euler_deg: Tuple[float, float, float] = (0, 0, 0),
    ):
        """
        initalize a transform

        ## Parameter
        - `vec_mm` : a 3D vector with unit in mm
        - `euler_deg` : a set of euler angle wiht unit in degree
        """
        self.vec_mm: Tuple[float, float, float] = vec_mm
        self.euler_deg: Tuple[float, float, float] = euler_deg

    def clone(self) -> "Transform":
        """clone the transform"""
        return Transform(vec_mm=self.vec_mm, euler_deg=self.euler_deg)

    def __repr__(self) -> str:
        return f"vec_mm : {self.vec_mm}, euler_deg : {self.euler_deg}"

    @classmethod
    def from_vec(cls, x_mm: float = 0, y_mm: float = 0, z_mm: float = 0) -> "Transform":
        """Construct a new `Transform` from vector componment"""
        return Transform(vec_mm=(x_mm, y_mm, z_mm))

    @classmethod
    def from_euler(
        cls, rx_deg: float = 0, ry_deg: float = 0, rz_deg: float = 0
    ) -> "Transform":
        """Construct a new `Transform` from euler angle set"""
        return Transform(euler_deg=(rx_deg, ry_deg, rz_deg))

    @classmethod
    def from_robot(cls, res: str) -> "Transform":
        """Parse robot transform response into `Transform`"""
        res = re.sub(r"[ {}]", "", res).split(",")
        t = Transform()
        for i in range(6):
            tokens = res[i].split(":")
            q = float(tokens[1])

            match tokens[0]:
                case "x":
                    t.set_x(q * 1000)
                case "y":
                    t.set_y(q * 1000)
                case "z":
                    t.set_z(q * 1000)
                case "rx":
                    t.set_rx(rad_to_deg(q))
                case "ry":
                    t.set_ry(rad_to_deg(q))
                case "rz":
                    t.set_rz(rad_to_deg(q))
                case _:
                    continue
        return t

    def vec_only(self) -> "Transform":
        """Extract translation from `self` and construct a new transfrom"""
        return Transform(vec_mm=self.vec_mm)

    def euler_only(self) -> "Transform":
        """Extract rotation from `self` and construct a new transfrom"""
        return Transform(euler_deg=self.euler_deg)

    @classmethod
    def from_x(cls, x_mm: float) -> "Transform":
        """Construct a new `Transform` with specified x"""
        return Transform.from_vec(x_mm=x_mm)

    @classmethod
    def from_y(cls, y_mm: float) -> "Transform":
        """Construct a new `Transform` with specified y"""
        return Transform.from_vec(y_mm=y_mm)

    @classmethod
    def from_z(cls, z_mm: float) -> "Transform":
        """Construct a new `Transform` with specified z"""
        return Transform.from_vec(z_mm=z_mm)

    @classmethod
    def from_rx(cls, rx_deg: float) -> "Transform":
        """Construct a new `Transform` with specified rx"""
        return Transform.from_euler(rx_deg=rx_deg)

    @classmethod
    def from_ry(cls, ry_deg: float) -> "Transform":
        """Construct a new `Transform` with specified ry"""
        return Transform.from_euler(ry_deg=ry_deg)

    @classmethod
    def from_rz(cls, rz_deg: float) -> "Transform":
        """Construct a new `Transform` with specified rz"""
        return Transform.from_euler(rz_deg=rz_deg)

    def set_vec(self, x_mm: float, y_mm: float, z_mm: float) -> "Transform":
        """set the vector"""
        self.vec_mm = (x_mm, y_mm, z_mm)
        return self

    def set_x(self, mm: float) -> "Transform":
        """set the x component to a specified value"""
        self.vec_mm = (mm, self.vec_mm[1], self.vec_mm[2])
        return self

    def set_y(self, mm: float) -> "Transform":
        """set the y component to a specified value"""
        self.vec_mm = (self.vec_mm[0], mm, self.vec_mm[2])
        return self

    def set_z(self, mm: float) -> "Transform":
        """set the z component to a specified value"""
        self.vec_mm = (self.vec_mm[0], self.vec_mm[1], mm)
        return self

    def set_euler(self, rx_deg: float, ry_deg: float, rz_deg: float) -> "Transform":
        """set the euler angle"""
        self.euler_deg = (rx_deg, ry_deg, rz_deg)
        return self

    def set_rx(self, deg: float) -> "Transform":
        """set the rx component to a specified value"""
        self.euler_deg = (deg, self.euler_deg[1], self.euler_deg[2])
        return self

    def set_ry(self, deg: float) -> "Transform":
        """set the ry component to a specified value"""
        self.euler_deg = (self.euler_deg[0], deg, self.euler_deg[2])
        return self

    def set_rz(self, deg: float) -> "Transform":
        """set the rz component to a specified value"""
        self.euler_deg = (self.euler_deg[0], self.euler_deg[1], deg)
        return self

    def to_homogenous(self) -> numpy_typing.NDArray[numpy.float64]:
        """
        return a homogenous matrix representation of the `self`

        ## Return:
        - `np.array` : 4x4 homogenous matrix representation of the transform
        """
        mat = numpy.eye(4)
        mat[0:3, 0:3] = euler_to_mat(self.euler_deg)
        mat[0:3, 3] = numpy.asarray(self.vec_mm)
        return mat

    @classmethod
    def from_homogenous(cls, mat: numpy_typing.NDArray[numpy.float64]) -> "Transform":
        """
        construct a transform from a `np.array` 4x4 homogenous matrix

        ## Parameter:
        - `mat : np.array` : 4x4 homogenous matrix representation of the transform

        ## Return:
        - `Transform`
        """
        return Transform(
            vec_mm=(
                mat[0, 3],
                mat[1, 3],
                mat[2, 3],
            ),
            euler_deg=mat_to_euler(mat),
        )

    def inv(self) -> "Transform":
        """return the inverse transform of `self`"""
        return Transform.from_homogenous(numpy.linalg.inv(self.to_homogenous()))

    def __mul__(self, rhs: "Transform") -> "Transform":
        return Transform.from_homogenous(
            numpy.matmul(self.to_homogenous(), rhs.to_homogenous())
        )

    def then(self, transform: "Transform") -> "Transform":
        """
        return a new transform that is apply a transform to `self`

        ## Parameter
        - `transform` : the transform to apply to `self`

        ## Return
        - `transform` : resulted transform
        """
        new = transform * self
        self.vec_mm = new.vec_mm
        self.euler_deg = new.euler_deg
        return self

    def then_x(self, cm: float) -> "Transform":
        """return a new transform that apply translation along x axis to `self`"""
        return self.then(Transform.from_x(cm))

    def then_y(self, cm: float) -> "Transform":
        """return a new transform that apply translation along y axis to `self`"""
        return self.then(Transform.from_y(cm))

    def then_z(self, cm: float) -> "Transform":
        """return a new transform that apply translation along z axis to `self`"""
        return self.then(Transform.from_z(cm))

    def then_rx(self, deg: float) -> "Transform":
        """return a new transform that apply rotation along x axis to `self`"""
        return self.then(Transform.from_rx(deg))

    def then_ry(self, deg: float) -> "Transform":
        """return a new transform that apply rotation along y axis to `self`"""
        return self.then(Transform.from_ry(deg))

    def then_rz(self, deg: float) -> "Transform":
        """return a new transform that apply rotation along z axis to `self`"""
        return self.then(Transform.from_rz(deg))

    def then_relative_to(
        self, transform: "Transform", reference: "Transform"
    ) -> "Transform":
        """
        return a new transform that,
        apply a transform to `self` relative to a reference

        ## Parameter
        - `transform` : the transform to apply
        - `reference` : the reference of the transfrom
        """
        return reference * transform * reference.inv() * self

    def then_relative(self, transform: "Transform") -> "Transform":
        """return a new transform that apply a transform relative `self`'s position"""
        return self.then_relative_to(transform=transform, reference=self.vec_only())

    def then_relative_rx(self, deg: float) -> "Transform":
        """return a new transform that apply relative rotaion along axis x"""
        return self.then_relative(Transform.from_rx(deg))

    def then_relative_ry(self, deg: float) -> "Transform":
        """return a new transform that apply relative rotaion along axis y"""
        return self.then_relative(Transform.from_ry(deg))

    def then_relative_rz(self, deg: float) -> "Transform":
        """return a new transform that apply relative rotaion along axis z"""
        return self.then_relative(Transform.from_rz(deg))

    def to_dict(self) -> dict[str, str | float]:
        """return a `dict[str,str|float]` representation of the transform"""
        return {
            "target": "transform",
            "x": self.vec_mm[0],
            "y": self.vec_mm[1],
            "z": self.vec_mm[2],
            "rx": self.euler_deg[0],
            "ry": self.euler_deg[1],
            "rz": self.euler_deg[2],
        }

    @classmethod
    def from_dict(cls, data: dict[str, str | float]) -> "Transform":
        """
        construct a new `Transform` from a dictionary

        ## Parameter
        - `data: dict[str, str|float]`: the data, if field is missing, 0 will be assumed
        """
        x = 0 if "x" not in data else data["x"]
        y = 0 if "y" not in data else data["y"]
        z = 0 if "z" not in data else data["z"]
        rx = 0 if "rx" not in data else data["rx"]
        ry = 0 if "ry" not in data else data["ry"]
        rz = 0 if "rz" not in data else data["rz"]

        return Transform((x, y, z), (rx, ry, rz))



"""
# Joint Coord Module
This module provide joint space representation class for joint space

## Class
- `JointCoord` : 6D joint space representation
"""

import re
from inovopy.geometry import IntoRobotCommand
from inovopy.geometry.transform import rad_to_deg


class JointCoord(IntoRobotCommand):
    """
    # JointCoord
    A class representing 6D joint space of robot
    """

    def __init__(
        self,
        j1_deg: float = 0,
        j2_deg: float = 0,
        j3_deg: float = 0,
        j4_deg: float = 0,
        j5_deg: float = 0,
        j6_deg: float = 0,
    ):
        """
        Construct a new joint space representation from all joint angle
        """
        self.joint_coord_deg: list[float] = [
            float(j1_deg),
            float(j2_deg),
            float(j3_deg),
            float(j4_deg),
            float(j5_deg),
            float(j6_deg),
        ]

    def __str__(self) -> str:
        return str([str(q) for q in self.joint_coord_deg])

    def __add__(self, rhs: "JointCoord") -> "JointCoord":
        out = JointCoord()
        out.joint_coord_deg = [(self[i] + rhs[i]) for i in range(6)]
        return out

    def __neg__(self) -> "JointCoord":
        neg = []

        for j in self.joint_coord_deg:
            neg.append(-j)

        JointCoord.from_list(neg)

    def __getitem__(self, idx: int) -> float:
        return self.joint_coord_deg[idx]

    @classmethod
    def from_robot(cls, res: str) -> "JointCoord":
        """
        Parse Robot input to `JointCoord`

        example input:

        `{joints : [-1.545941, -0.087197, -0.111058, -2.998237, 1.762412, 3.399661, ],
        tcp : {rx : -1.793361, ry : 0.255386, rz : 1.682603,
        x : -0.073378, y : -0.014815, z : 0.929764, },
        tcpid : tool_plate, }`
        """
        res = re.findall(r"\[.*\]", res)[0]
        res = re.sub(r"[ \[\]]", "", res).split(",")
        joints: list[float] = []
        for i in range(6):
            try:
                joints.append(rad_to_deg(float(res[i])))
            except (ValueError, IndexError):
                joints.append(0)
        return JointCoord(
            joints[0],
            joints[1],
            joints[2],
            joints[3],
            joints[4],
            joints[5],
        )

    @classmethod
    def from_list(cls, coord: list[float]) -> "JointCoord":
        """Construst a new `JointCoord` with specified coord"""
        assert len(coord) >= 6
        for i in range(6):
            coord[i] = float(coord[i])
        return JointCoord(
            j1_deg=coord[0],
            j2_deg=coord[1],
            j3_deg=coord[2],
            j4_deg=coord[3],
            j5_deg=coord[4],
            j6_deg=coord[5],
        )

    @classmethod
    def from_j1(cls, deg: float = 0) -> "JointCoord":
        """Construct a new `JointCoord` with specified joint 1"""
        return JointCoord(j1_deg=deg)

    @classmethod
    def from_j2(cls, deg: float = 0) -> "JointCoord":
        """Construct a new `JointCoord` with specified joint 2"""
        return JointCoord(j1_deg=deg)

    @classmethod
    def from_j3(cls, deg: float = 0) -> "JointCoord":
        """Construct a new `JointCoord` with specified joint 3"""
        return JointCoord(j1_deg=deg)

    @classmethod
    def from_j4(cls, deg: float = 0) -> "JointCoord":
        """Construct a new `JointCoord` with specified joint 4"""
        return JointCoord(j1_deg=deg)

    @classmethod
    def from_j5(cls, deg: float = 0) -> "JointCoord":
        """Construct a new `JointCoord` with specified joint 5"""
        return JointCoord(j1_deg=deg)

    @classmethod
    def from_j6(cls, deg: float = 0) -> "JointCoord":
        """Construct a new `JointCoord` with specified joint 6"""
        return JointCoord(j1_deg=deg)

    def then_j1(self, deg: float = 0) -> "JointCoord":
        """Return a new `JointCoord` with a by roating joint 1"""
        return self + JointCoord.from_j1(deg)

    def then_j2(self, deg: float = 0) -> "JointCoord":
        """Return a new `JointCoord` with a by roating joint 2"""
        return self + JointCoord.from_j2(deg)

    def then_j3(self, deg: float = 0) -> "JointCoord":
        """Return a new `JointCoord` with a by roating joint 3"""
        return self + JointCoord.from_j3(deg)

    def then_j4(self, deg: float = 0) -> "JointCoord":
        """Return a new `JointCoord` with a by roating joint 4"""
        return self + JointCoord.from_j4(deg)

    def then_j5(self, deg: float = 0) -> "JointCoord":
        """Return a new `JointCoord` with a by roating joint 5"""
        return self + JointCoord.from_j5(deg)

    def then_j6(self, deg: float = 0) -> "JointCoord":
        """Return a new `JointCoord` with a by roating joint 6"""
        return self + JointCoord.from_j6(deg)

    def to_dict(self) -> dict[str, str | float]:
        """return a `dict[str,str|float]` representation of the `JointCoord`"""
        return {
            "target": "joint_coord",
            "j1": self.joint_coord_deg[0],
            "j2": self.joint_coord_deg[1],
            "j3": self.joint_coord_deg[2],
            "j4": self.joint_coord_deg[3],
            "j5": self.joint_coord_deg[4],
            "j6": self.joint_coord_deg[5],
        }

    @classmethod
    def from_dict(cls, data: dict[str, str | float]) -> "JointCoord":
        """
        construct a new `JointCoord` from a dictionary

        ## Parameter
        - `data: dict[str, str|float]`: the data, if field is missing, 0 will be assumed
        """
        j1 = 0 if "j1" not in data else data["j1"]
        j2 = 0 if "j2" not in data else data["j2"]
        j3 = 0 if "j3" not in data else data["j3"]
        j4 = 0 if "j4" not in data else data["j4"]
        j5 = 0 if "j5" not in data else data["j5"]
        j6 = 0 if "j6" not in data else data["j6"]
        return JointCoord(j1, j2, j3, j4, j5, j6)



"""
# TCP Listener Module
This module provide an api class for listening tcp connection

## Classes
- `TcpListener` : a class for managing and listening connection
"""

import socket
import select
import logging

from inovopy.socket import SocketException, detect_interfaces
from inovopy.socket.tcp_stream import TcpStream
from inovopy.util import Loggable

DEFAULT_PORT: int = 50003
"""Default port to listen on if no poort are specified `50003`."""


class TcpListener(Loggable):
    """
    # TcpListener
    A class for listening to tcp connection

    ## Usage
    ```python
    from inovopy.socket import TcpListener

    example_listener = TcpListener()

    example_stream = example_listener.accept()
    ```
    """

    def __init__(
        self,
        host: str | None = None,
        port: int = DEFAULT_PORT,
        logger: logging.Logger | str | None = None,
    ):
        """
        initalize the tcp listener

        do not specify the host if you wish to listen at all
        of the host in your machine.

        specify the host if you wish to listen at a specific
        network in you machine

        ## Parameter
        - `host : str | None` : local ip
        - `port : int` : port to listen to
        - `logger: logging.Logger | str | None` :
            - if `logger` is instance of `logging.Logger`, it will log with it;
            - if `logger` is `str`, new logger will be created with it as name
            - otherwise, no log

        ## Exception:
        `SocketException`:
        - if no network interface found
        - if not vaild socket are created
        """
        super().__init__(logger)

        self.debug("creating TcpListener . . .")

        if host is None:
            self.info("host not supplied, detecting networks . . .")
            ips = detect_interfaces()
            """a list of ip to bind socket to and listen to"""

            if len(ips) == 0:
                self.error("No interface found")
                raise SocketException(
                    "No Network Interfaces, please make sure device is connected to network"
                )
        else:
            ips = [host]

        self.info("creating sockets . . .")

        self.__sockets: list[socket.socket] = []
        """a list of socket to listen to"""

        for ip in ips:
            self.debug(f"Listening @ ip {ip}")
            soc: socket.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

            self.debug("....setting socket option")
            soc.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

            self.debug("....trying to bind . . .")
            try:
                soc.bind((ip, port))
                self.debug("....socket binding successful.")
            except OSError as e:
                self.warning(f"Failed to bind @ {ip}:{port}")
                self.warning(f"Error: {e}")
                continue

            self.debug("....trying to listen . . .")
            try:
                soc.listen()
                self.debug("....socket start listening . . .")
            except OSError as e:
                self.warning(f"Failed to listen @ {ip}:{port}")
                self.warning(f"{e}")
                continue

            soc.setblocking(False)

            self.info(f"Finished setting up socket @ {ip}:{port}")
            self.__sockets.append(soc)

        if len(self.__sockets) == 0:
            self.error("No vaild socket created.")
            raise SocketException("No vaild socket created.")

    def accept(self, stream_logger: logging.Logger | str | None = None) -> TcpStream:
        """
        try accept a new tcp connection.

        ## Parameter:
        - `logger: logging.Logger | str | None` : logger of returned `TcpStream`
            - if `logger` is instance of `logging.Logger`, it will log with it;
            - if `logger` is `str`, new logger will be created with it as name
            - otherwise, no log

        ## Return:
        `TcpStream` : the accepted tcp connetion

        ## Exception:
        No exception, will continue to try to accept if encounter `OSError`
        """
        self.info("Start accepting . . .")

        self.debug("waiting for connection . . .")
        while True:
            readables, _, _ = select.select(self.__sockets, [], [], 1)

            for soc in readables:
                soc: socket.socket = soc

                if soc not in self.__sockets:
                    self.warning("....readables not in self.__sockets")
                    continue

                (ip, port) = soc.getsockname()

                self.info(f"trying to accept connection at {ip}:{port}")

                try:
                    accept = soc.accept()
                    conn: socket.socket = accept[0]
                    peer_ip, peer_port = accept[1]
                    self.info(f"Accepting connect successful to {peer_ip}:{peer_port}")
                except OSError as e:
                    self.warning("Accepting connection failed")
                    self.warning(f"{e}")
                    continue

                stream = TcpStream(conn, stream_logger)
                return stream

    def __del__(self):
        for soc in self.__sockets:
            try:
                (ip, port) = soc.getsockname()
                self.debug(f"shutting down socket @ {ip}:{port}")
                soc.close()
            except OSError as e:
                self.error("Failed to shut down socket!")
                self.error(f"{e}")
                continue



"""
# inovopy
This library provide simple api for application interacting with
inovo robot arm.

## Modules
- `socket`, tcp socket-based communication api
- `geometry`, joint-space and 3D spatial data structure and api
- `iva`, message generation for iva protocal (a robot protocal interfacing with inovo robot arm)
- `rosbridge`, rosbridge API to control psu
- `util`, utilility module
"""



"""
# Utils Module
This module provide utility functions and classes.

## Classes
`ConsoleHandler`: colorful python logging handler
`Loggable` : helper interface for logging class

## Functions
- `clean` : clean up all non visible character of a str
- `clamp` : clamp float
"""

import logging
import abc
import re


def clean(s: str) -> str:
    """clean up all non visible characters of a str"""
    return re.sub(r"[^ -~]", "", s)


def clamp(f: float, floor: float, ceil: float) -> float:
    """clamp a float between a floor and a ceil"""
    return min(max(f, floor), ceil)


class ConsoleHandler(logging.Handler):
    """
    A class for logging to console with color

    Base Class:
    - `logging.Handler`
    """

    class ConsoleColor:
        """
        a internal class for changing console display color
        """

        RED = "\033[31m"
        YELLOW = "\033[33m"
        GREEN = "\033[32m"
        RESET = "\033[0m"

    def __init__(self, level=0):
        super().__init__(level)

    def emit(self, record: logging.LogRecord):
        cc = ConsoleHandler.ConsoleColor

        l = record.levelno
        msg = self.format(record)

        if l >= 40:
            msg = f"{cc.RED}{msg}{cc.RESET}"
        elif l >= 30:
            msg = f"{cc.YELLOW}{msg}{cc.RESET}"
        elif l >= 20:
            msg = f"{cc.GREEN}{msg}{cc.RESET}"
        else:
            pass

        print(msg)


def get_package_logger() -> logging.Logger:
    """Get package logger"""
    return logging.getLogger("inovopy")


class Loggable(abc.ABC):
    """
    A simple interface baseclass for logging class

    Function:
    - `debug(*args, **kwargs)` : transparent to `logger.Logger.debug`
    - `info(*args, **kwargs)` : transparent to `logger.Logger.info`
    - `warning(*args, **kwargs)` : transparent to `logger.Logger.warning`
    - `error(*args, **kwargs)` : transparent to `logger.Logger.error`
    - `critical(*args, **kwargs)` : transparent to `logger.Logger.critical`
    """

    logger: logging.Logger | None = None

    @classmethod
    def get_class_logger(cls) -> logging.Logger:
        return get_package_logger().getChild(cls.__name__)

    @classmethod
    def get_logger(cls, name: str) -> logging.Logger:
        return cls.get_class_logger().getChild(name)

    def __init__(self, logger: logging.Logger | str | None = None):
        super().__init__()

        if isinstance(logger, logging.Logger):
            self.logger = logger
        elif isinstance(logger, str):
            self.logger = self.get_logger(logger)

    def debug(self, *args, **kwargs):
        """transparent to `logger.Logger.debug`"""
        if self.logger:
            self.logger.debug(*args, **kwargs)

    def info(self, *args, **kwargs):
        """transparent to `logger.Logger.info`"""
        if self.logger:
            self.logger.info(*args, **kwargs)

    def warning(self, *args, **kwargs):
        """transparent to `logger.Logger.warning`"""
        if self.logger:
            self.logger.warning(*args, **kwargs)

    def error(self, *args, **kwargs):
        """transparent to `logger.Logger.error`"""
        if self.logger:
            self.logger.error(*args, **kwargs)

    def critical(self, *args, **kwargs):
        """transparent to `logger.Logger.critical`"""
        if self.logger:
            self.logger.critical(*args, **kwargs)


LOGGING_CONFIG_CONSOLE = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "simpleformatter": {
            "format": "%(name)-10s %(levelname)-7s %(message)s",
        },
    },
    "handlers": {
        "consolehandler": {
            "()": "inovopy.util.ConsoleHandler",
            "formatter": "simpleformatter",
            "level": "INFO",
        },
    },
    "loggers": {
        "root": {
            "handlers": ["consolehandler"],
            "level": "DEBUG",
            "propagate": True,
        }
    },
}
"""
A simple config for `logging`, console only

Usage:
```python
from logging.config import dictConfig
from inovopy.util import LOGGING_CONFIG_CONSOLE

dictConfig(LOGGING_CONFIG_CONSOLE)
```
"""

LOGGING_CONFIG = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "simpleformatter": {
            "format": "%(name)-10s %(levelname)-7s %(message)s",
        },
        "detailformatter": {
            "format": "%(asctime)s %(name)-10s %(levelname)-7s %(funcName)s() L%(lineno)-4d %(message)s",
            "datefmt": "%Y-%m-%d %H:%M:%S",
        },
    },
    "handlers": {
        "rotatehandler": {
            "class": "logging.handlers.RotatingFileHandler",
            "formatter": "detailformatter",
            "filename": "logging/master_log.log",
            "maxBytes": 50 * 1000 * 1000,
            "backupCount": 3,
        },
        "consolehandler": {
            "()": "inovopy.util.ConsoleHandler",
            "formatter": "simpleformatter",
            "level": "INFO",
        },
    },
    "loggers": {
        "root": {
            "handlers": ["consolehandler", "rotatehandler"],
            "level": "DEBUG",
            "propagate": True,
        }
    },
}
"""
A simple config for `logging`, console + rotating file

Usage:
```python
from logging.config import dictConfig
from inovopy.util import LOGGING_CONFIG

dictConfig(LOGGING_CONFIG)
```
"""



"""
# Geometry Module
This module provide geometric data structure class for robot control

## Modules
- `transform` : Functions and Class for 3D spatial transform
- `jointcoord` : 6D joint space representation
"""

import numpy
from abc import ABC
from inovopy.iva import MotionMode, RobotCommand


def deg_to_rad(deg: float) -> float:
    """translate degree to radian"""
    return deg / 180 * numpy.pi


def rad_to_deg(rad: float) -> float:
    """translate radion to degree"""
    return rad / numpy.pi * 180


class IntoRobotCommand(ABC):
    """
    # IntoRobotCommand
    An interface for all class that can be turn into robot command

    ## Method
    - `as_motion`
    - `as_linear`
    - `as_linear_relative`
    - `as_joint`
    - `as_joint_relative`
    """

    # @abstractmethod
    def as_motion(self, motion_mode: MotionMode) -> RobotCommand:
        """construct a new motion command from the `self`"""
        return RobotCommand.motion(motion_mode=motion_mode, target=self)

    # @abstractmethod
    def as_linear(self) -> RobotCommand:
        """construct a new linear motion command from the `self`"""
        return self.as_motion(motion_mode=MotionMode.LINEAR)

    # @abstractmethod
    def as_linear_relative(self) -> RobotCommand:
        """construct a new linear relative motion command from the `self`"""
        return self.as_motion(motion_mode=MotionMode.LINEAR_RELATIVE)

    # @abstractmethod
    def as_joint(self) -> RobotCommand:
        """construct a new joint motion command from the `self`"""
        return self.as_motion(motion_mode=MotionMode.JOINT)

    # @abstractmethod
    def as_joint_relative(self) -> RobotCommand:
        """construct a new joint relative motion command from the `self`"""
        return self.as_motion(motion_mode=MotionMode.JOINT_RELATIVE)


from inovopy.geometry.transform import Transform
from inovopy.geometry.jointcoord import JointCoord



"""
# IVA Module
This module provide class and function for
message generation of IVA communication protocal to inovo robots

## Functions
- `execute` : generate `dict` for `execute` message for robot
- `enqueue` : generate `dict` for `enqueue` message for robot
- `dequeue` : generate `dict` for `dequeue` message for robot
- `pop` : generate `dict` for `pop` message for robot
- `io` : generate `dict` for `io` message for robot
- `gripper` : generate `dict` for `gripper` message for robot
- `get_current` : generate `dict` for `get` current message for robot
- `get_data` : generate `dict` for `get` data message for robot
- `custom` : generate `dict` for `custom` message for robot

## Class
- `RobotCommand` ; class for constructing robot command
- `IOCommand` ; class for constructing IO command
- `GripperCommand` ; class for constructing gripper command

## Interface, Abstract Base Class
- `IntoRobotCommand` : interface for class to constructing robot command
"""

from typing import Literal
from enum import Enum
import numpy
from inovopy.util import clamp
import inovopy


def execute(
    robot_command: "RobotCommand", enter_context: bool = False
) -> dict[str, str | float]:
    """
    generate jsonable `dict[str,str|float]` for execute command

    ## Parameter
    - `robot_command : RobotCommand` : the robot command to execute
    - `enter_context : bool` : whether or not to push to the context stack
    """
    return {
        "op_code": "execute",
        "enter_context": 1 if enter_context else 0,
        **robot_command.to_dict(),
    }


def enqueue(robot_command: "RobotCommand") -> dict[str, str | float]:
    """
    generate jsonable `dict[str,str|float]` for enqueue command

    ## Parameter
    - `robot_command : RobotCommand` : the robot command to enqueue
    """
    return {"op_code": "enqueue", **robot_command.to_dict()}


def dequeue(enter_context: bool = False) -> dict[str, str | float]:
    """
    generate jsonable `dict[str,str|float]` for dequeue command

    ## Parameter
    - `enter_context : bool` : whether or not to push to the context stack
    """
    return {"op_code": "dequeue", "enter_context": 1 if enter_context else 0}


def pop() -> dict[str, str | float]:
    """
    generate jsonable `dict[str,str|float]` for pop command
    """
    return {
        "op_code": "pop",
    }


def gripper(gripper_command: "GripperCommand") -> dict[str, str | float]:
    """
    generate jsonable `dict[str,str|float]` for gripper command

    ## Parameter
    - `gripper_command : GripperCommand` : gripper command to execute
    """
    return {"op_code": "gripper", **gripper_command.to_dict()}


def io(io_command: "IOCommand") -> dict[str, str | float]:
    """
    generate jsonable `dict[str,str|float]` for io command

    ## Parameter
    - `io_command : IOCommand` : io command to execute
    """
    return {"op_code": "io", **io_command.to_dict()}


def get_current(target: Literal["transform", "joint_coord"]) -> dict[str, str | float]:
    """
    generate jsonable `dict[str,str|float]` for get command

    ## Parameter
    - `transform : bool` : whether to get the current transform or joint
    """
    return {
        "op_code": "get",
        "target": "transform" if target == "transform" else "joint_coord",
    }


def get_data(key: str) -> dict[str, str | float]:
    """
    generate jsonable `dict[str,str|float]` for get data command

    ## Parameter
    - `key : str` : key of the data to get from the robot
    """
    return {"op_code": "get", "target": "data", "key": key}


def custom(custom_command: dict[str, str | float]) -> dict[str, str | float]:
    """
    generate jsonable `dict[str,str|float]` for custom command

    ## Parameter
    - `custom_command : dict[str,str|float]` : custom command to execute
    """
    return {"op_code": "custom", **custom_command}


class RobotCommand:
    """
    # RobotCommand
    a class for constructing robot command

    ## Class Method
    - `synchronize` : command to break blending
    - `sleep` : command to robot to sleep
    - `set_parameter` : command to set motion parameter of robot
    - `motion` : command to move the robot
    """

    def __init__(self, argument: dict[str, str | float]):
        self.argument: dict[str, str | float] = argument

    def to_dict(self) -> dict[str, str | float]:
        """return `dict[str,str|float]` representation of the command"""
        return self.argument

    @classmethod
    def synchronize(cls) -> "RobotCommand":
        """Return a `RobotCommand` for synchronizing/breaking blending"""
        return RobotCommand(
            {
                "action": "synchronize",
            }
        )

    @classmethod
    def sleep(cls, second: float) -> "RobotCommand":
        """
        Return a `RobotCommand` for the robot to sleep

        ## Parameter
        - `second : float` : specify the lenght of sleep in seconds
        """
        return RobotCommand({"action": "sleep", "second": second})

    @classmethod
    def set_parameter(
        cls,
        speed: float | None = None,
        accel: float | None = None,
        blend_linear: float | None = None,
        blend_angular: float | None = None,
        tcp_speed_linear: float | None = None,
        tcp_speed_angular: float | None = None,
    ) -> "RobotCommand":
        """
        Return a `RobotCommand` for setting the robot motion parameter

        ## Parameter
        - `speed : float`, in percent, range from `1` to `100`
        - `accel : float`, in precent, range from `1` to `100`
        - `blend_linear : float`, in mm, range from `1` to `1000`
        - `blend_angular : float`, in degree, range from `1` to `360`
        - `tcp_speed_linear : float`, in mm, range from `1` to `999`
        - `tcp_speed_angular : float`, in degree, range from `1` to `360`
        """

        speed = clamp(speed / 100, 0.01, 1) if speed else 0

        accel = clamp(accel / 100, 0.01, 1) if accel else 0

        blend_linear = clamp(blend_linear / 1000, 0.001, 1) if blend_linear else 0

        blend_angular = (
            clamp(
                inovopy.geometry.transform.deg_to_rad(blend_angular),
                0.001,
                2 * numpy.pi,
            )
            if blend_angular
            else 0
        )

        tcp_speed_linear = (
            clamp(tcp_speed_linear / 1000, 0.001, 0.999) if tcp_speed_linear else 0
        )

        tcp_speed_angular = (
            clamp(
                inovopy.geometry.transform.deg_to_rad(tcp_speed_angular),
                0.001,
                2 * numpy.pi,
            )
            if tcp_speed_angular
            else 0
        )

        return RobotCommand(
            {
                "action": "set_parameter",
                "speed": speed,
                "accel": accel,
                "blend_linear": blend_linear,
                "blend_angular": blend_angular,
                "tcp_speed_linear": tcp_speed_linear,
                "tcp_speed_angular": tcp_speed_angular,
            }
        )

    @classmethod
    def motion(
        cls,
        motion_mode: "MotionMode",
        target: "inovopy.geometry.jointcoord.JointCoord | inovopy.geometry.transform.Transform",
    ) -> "RobotCommand":
        """
        Return a `RobotCommand` for robot motion

        ## Parameter
        - `motion_mode : MotionMode`, specify the motion mode
        - `target` : target of the motion, either transform or joint coord
        """
        return RobotCommand(
            {
                "action": "motion",
                "motion_mode": motion_mode.to_arg(),
                **target.to_dict(),
            }
        )


class MotionMode(str, Enum):
    """
    # MotionMode
    An str-enum class for different robot motion's
    interpolation mode and specify whether is relative move

    ## Enum
    - `LINEAR`
    - `LINEAR_RELATIVE`
    - `JOINT`
    - `JOINT_RELATIVE`
    """

    LINEAR = "linear"
    LINEAR_RELATIVE = "linear_relative"
    JOINT = "joint"
    JOINT_RELATIVE = "joint_relatve"

    def to_arg(self) -> str:
        return str(self).split(".")[1].lower()


class IOCommand:
    """
    # IO Command
    A class for constructing IO Command

    ## Class Method
    - `set_beckhoff`
    - `get_beckhoff`
    - `set_wrist`
    - `get_wrist`
    """

    def __init__(self, argument: dict[str, str | float]):
        self.argument = argument

    def to_dict(self) -> dict[str, str | float]:
        """return a `dict[str,str|float]` representation of the io command"""
        return self.argument

    @classmethod
    def set_digital(
        cls, target: Literal["beckhoff", "wrist"], port: int, state: bool
    ) -> "IOCommand":
        """
        return a command for setting the digital output

        ## Parameter
        - `port : int` : port to set
        - `state : bool` : state to set
        """
        return IOCommand(
            {
                "action": "set",
                "target": "beckhoff" if target == "beckhoff" else "wrist",
                "port": port,
                "state": 1 if state else 0,
            }
        )

    @classmethod
    def get_digital(
        cls, target: Literal["beckhoff", "wrist"], port: int
    ) -> "IOCommand":
        """
        return a command for getting the digital input

        ## Parameter
        - `port : int` : port to set
        """
        return IOCommand(
            {
                "action": "get",
                "target": "beckhoff" if target == "beckhoff" else "wrist",
                "port": port,
            }
        )


class GripperCommand:
    """
    # Gripper Command
    A class for constructing gripper command

    ## Class Method
    - `activate`
    - `set`
    - `get`
    """

    def __init__(self, argument: dict[str, str | float]):
        self.argument = argument

    def to_dict(self) -> dict[str, str | float]:
        """return `dict[str,str|float]` representation of the command"""
        return self.argument

    @classmethod
    def activate(cls) -> "GripperCommand":
        """return a gripper command for activating the gripper"""
        return GripperCommand({"action": "activate"})

    @classmethod
    def set(cls, label: str) -> "GripperCommand":
        """
        return a gripper command for setting the gripper

        # Parameter
        - `label : str` : the label to set the gripper to
        """
        return GripperCommand({"action": "set", "label": label})

    @classmethod
    def get(cls) -> "GripperCommand":
        """return a gripper command for getting the state of the robot"""
        return GripperCommand({"action": "get"})



"""
# Socket Module
This module provide api for handle generic socket communication
for communicating with robot for control.

## Class
- `TcpListener` : A class for managing listening socket connection
- `TcpStream` : A class for managing socket communication to a client
- `SocketExecption` : An Excpetion Class for all socket related exception
## Example

```python
from inovopy.socket import TcpListener, TcpStream

listener = TcpListener()               # <-- create a new listener
stream: TcpStream = listener.accept()  # <-- accept a new connection
del(listener)                          # <-- delete the listener to stop listening

stream.write("Send this message")      # <-- send a str message
read_message : str = stream.read()     # <-- read a messagea as str
```
"""

import socket
import platform


class SocketException(Exception):
    """socket commuication exception"""


class EndOfCommunication(Exception):
    """end of communication exception"""


def detect_interfaces() -> list[str]:
    """
    try to automatically detect local ip of this machine
    in all local networks

    ## Return:
    `list[str]` a list of ip
    """
    hostname = socket.gethostname()
    if platform.system() == "Linux":
        hostname += ".local"
    return socket.gethostbyname_ex(hostname)[2]


from inovopy.socket.tcp_listener import TcpListener
from inovopy.socket.tcp_stream import TcpStream



"""
setup script for inovopy
"""

from setuptools import setup, find_packages
from pathlib import Path

VERSION = "1.0.0"
DESCRIPTION = "Inovo Robot Arm API"
LONG_DESCRIPTION = (Path(__file__).parent / "README.md").read_text()

# Setting up
setup(
    name="inovopy",
    version=VERSION,
    author="Alan Chung",
    author_email="deeralan827@gmail.com",
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    install_requires=["numpy", "roslibpy"],
    keywords=[
        "python",
        "robotics",
        "inovo robotics",
        "inovo robot arm",
        "motion",
        "sockets",
    ],
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ],
)



"""
# Robot Module
This module provide a simple api for controlling inovo robot arm with
predefinded `iva` protocal.

## Class
- `InovoRobot`: A class for interfacing with inovo robot arm

## Example
```python
from inovopy.robot import InovoRobot

bot = InovoRobot.default_iva("psu000")


```
"""

import logging
from contextlib import contextmanager
import json

from inovopy.util import Loggable
from inovopy.rosbridge import InovoRos
from inovopy.socket import TcpListener, TcpStream
from inovopy.geometry.jointcoord import JointCoord
from inovopy.geometry.transform import Transform
from inovopy.iva import *


class IvaException(Exception):
    """
    # Iva Exception
    """


class InovoRobot(Loggable):
    """
    # InovoRobot
    An API class for controlling and access information from inovo robot arm system.

    # Usage
    ```python
    from inovopy.robot import InovoRobot

    bot = InovoRobot.default_iva("psu000")

    print(bot.get_current_transform())
    print(bot.get_current_jointcoord())
    ```
    """

    def __init__(
        self,
        tcp_stream: TcpStream,
        inovo_ros: InovoRos,
        logger: logging.Logger | str | None = None,
    ):
        super().__init__(logger)
        self.tcp_stream: TcpStream = tcp_stream
        self.inovo_ros: InovoRos = inovo_ros

    @classmethod
    def default_iva(
        cls, host: str, logger: logging.Logger | str | None = None
    ) -> "InovoRobot":
        """
        start the iva protocal by

        - start up a tcp listener
        - start a sequence on psu via rosbridge
        - accept a connection

        ## Parameter:
        - `host: str` : remote host to start sequence, aka psu's address

        ## Return:
        `InovoRobot` : the resulted api class
        """
        listener = TcpListener()

        inovo_ros: InovoRos = InovoRos(host=host)
        inovo_ros.runtime_start("iva")

        tcp_stream: TcpStream = listener.accept()
        del listener

        bot: InovoRobot = InovoRobot(
            tcp_stream=tcp_stream, inovo_ros=inovo_ros, logger=logger
        )
        return bot

    def read(self) -> str:
        """read a line"""
        return self.tcp_stream.read()

    def write(self, instruction: dict[str, str | float]):
        """
        write an instruction as str

        ## Parameter:
        - `instruction : dict[str, str|float]` : jsonable instruction dict
        """
        msg = json.dumps(instruction)
        self.tcp_stream.write(msg)

    def assert_res_ok(self):
        """read a message and assert that it is `OK`"""
        res = self.read()
        if res != "OK":
            raise IvaException(f'Expect resopnse to be "OK", but recieve {res}')

    def execute(self, robot_command: RobotCommand, enter_context: bool = False):
        """
        instruct the robot to execute a `inovopy.iva.RobotCommand`

        ## Parameter:
        - `robot_command: inovopy.iva.RobotCommand`: command to execute
        - `enter_context: bool`: whether the execution of command create a pop-able context

        ## Exception:
        `IvaException` raise if response is not `OK`
        """
        self.write(execute(robot_command=robot_command, enter_context=enter_context))
        self.assert_res_ok()

    def sleep(self, second: float):
        """
        instruct the robot to sleep for a specified time

        ## Parameter:
        - `second: float` : second to sleep
        """
        self.execute(RobotCommand.sleep(second=second))

    def set_param(
        self,
        speed: float | None = None,
        accel: float | None = None,
        blend_linear: float | None = None,
        blend_angular: float | None = None,
        tcp_speed_linear: float | None = None,
        tcp_speed_angular: float | None = None,
    ):
        """
        set the motion parameter of the robot

        ## Parameter
        - `speed : float`, in percent, range from `1` to `100`
        - `accel : float`, in precent, range from `1` to `100`
        - `blend_linear : float`, in mm, range from `1` to `1000`
        - `blend_angular : float`, in degree, range from `1` to `360`
        - `tcp_speed_linear : float`, in mm, range from `1` to `999`
        - `tcp_speed_angular : float`, in degree, range from `1` to `360`
        """
        robot_command = RobotCommand.set_parameter(
            speed=speed,
            accel=accel,
            blend_linear=blend_linear,
            blend_angular=blend_angular,
            tcp_speed_linear=tcp_speed_linear,
            tcp_speed_angular=tcp_speed_angular,
        )

        self.execute(robot_command=robot_command)

    def linear(self, target: Transform | JointCoord):
        """linear move to a target"""
        self.execute(target.as_linear())

    def linear_relative(self, target: Transform | JointCoord):
        """linear relative move to a target"""
        self.execute(target.as_linear_relative())

    def joint(self, target: Transform | JointCoord):
        """joint move to a target"""
        self.execute(target.as_joint())

    def joint_relative(self, target: Transform | JointCoord):
        """joint relative move to a target"""
        self.execute(target.as_joint_relative())

    def enqueue(self, robot_command: RobotCommand):
        """
        instruct the robot to enqueue a command

        ## Parameter
        - `robot_command: inovopy.iva.RobotCommand`: the command to enqueue
        """
        self.write(enqueue(robot_command=robot_command))
        self.assert_res_ok()

    def dequeue(self, enter_context: bool = False):
        """
        instruct the robot to dequeue all enqueued command

        ## Parameter
        - `enter_context: bool`: whether enter a context with the sequence or not
        """
        self.write(dequeue(enter_context=enter_context))
        self.assert_res_ok()

    def sequence(self, sequence: list[RobotCommand], enter_context: bool = False):
        """
        perform a sequence of `inovopy.iva.RobotCommand`

        ## Parameter
        - `enter_context: bool`: whether enter a context with the sequence or not
        """
        for step in sequence:
            self.enqueue(step)
        self.dequeue(enter_context=enter_context)

    def pop(self):
        """
        exit a context
        """
        self.write(pop())
        self.assert_res_ok()

    def io(self, io_command: IOCommand):
        """
        instruct the robot to perform a `inovopy.iva.IOCommand`

        ## Parameter
        - `io_command: inovopy.iva.IOCommand`: io commmand to perform

        ## Return
        `str` the respons
        """
        self.write(io(io_command=io_command))
        return self.read()

    def get_io_beckhoff(self, port: int) -> bool:
        """
        get beckhoff input port

        ## Parameter
        - `port: int`: input port to get, default `0-7`

        ## Return
        `bool` state of io
        """
        return self.io(IOCommand.get_digital(target="beckhoff", port=port)) == "True"

    def get_io_wrist(self, port: int) -> bool:
        """
        get wrist input port

        ## Parameter
        - `port: int`: input port to get, default `0-1`

        ## Return
        `bool` state of io
        """
        return self.io(IOCommand.get_digital(target="wrist", port=port)) == "True"

    def set_io_beckhoff(self, port: int, state: bool):
        """
        set beckhoff output port

        ## Parameter
        - `port: int`: output port to set, default `0-7`
        - `state: bool`: target state of the output
        """
        res = self.io(IOCommand.set_digital(target="beckhoff", port=port, state=state))
        assert res == "OK"

    def set_io_wrist(self, port: int, state: bool):
        """
        set wrist output port

        ## Parameter
        - `port: int`: output port to set, default `0-7`
        - `state: bool`: target state of the output
        """
        res = self.io(IOCommand.set_digital(target="wrist", port=port, state=state))
        assert res == "OK"

    def gripper(self, gripper_command: GripperCommand):
        self.write(gripper(gripper_command))

    def gipper_activate(self):
        """activate the gripper"""
        self.gripper(GripperCommand.activate())
        self.assert_res_ok()

    def gripper_get(self) -> float:
        """
        get the gripper width

        ## Return
        `float` in percentage
        """
        self.gripper(GripperCommand.get())
        return float(self.read()) * 100

    def gripper_set(self, label: str):
        """
        set a gripper to a predefine label


        #Parameter
        - `label: str`: the label to set to
        """
        self.gripper(GripperCommand.set(label=label))
        self.assert_res_ok()

    def get_current_transform(self) -> Transform:
        """
        get the current transform

        ## Return:
        `inovopy.geometry.transform.Transform` : current tranform
        """
        self.write(get_current("transform"))
        return Transform.from_robot(self.read())

    def get_current_joint(self) -> JointCoord:
        """
        get the current joint coordinate

        ## Return:
        `inovopy.geometry.jointcoord.JointCoord`: current joint coordinate
        """
        self.write(get_current("joint_coord"))
        return JointCoord.from_robot(self.read())

    def get_data(self, key: str) -> str:
        """
        get data with a key

        ## Parameter:
        - `key : str` : key for the data

        ## Return:
        - `str` value of the data
        - `str` containing `Error` will return if the key is not found
        """
        self.write(get_data(key=key))
        return self.read()

    @contextmanager
    def context(self, robot_command: RobotCommand):
        """
        enter a context with a robot command

        ## Parameter
        - `robot_command: inovopy.iva.RobotCommand` : the command to execute

        ## Context Manager
        when exit context `InovoRobot.pop` the context
        """
        self.execute(robot_command=robot_command, enter_context=True)
        try:
            yield
        finally:
            self.pop()

    @contextmanager
    def context_sequence(self, sequence: list[RobotCommand]):
        """
        enter a context with a robot command

        ## Parameter
        - `sequence: list[inovopy.iva.RobotCommand]` : the sequence to execute

        ## Context Manager
        when exit context `InovoRobot.pop` the context
        """
        self.sequence(sequence=sequence, enter_context=True)
        try:
            yield
        finally:
            self.pop()
